DROP FUNCTION IF EXISTS proc_events_delete(regclass, regclass, character varying(255));
DROP FUNCTION IF EXISTS proc_events_delete_with_no_parent_id_constraint(regclass, character varying(255), character varying(255));
